#ifndef __FWEELIN_FLUIDSYNTH_H
#define __FWEELIN_FLUIDSYNTH_H

#if USE_FLUIDSYNTH

#include <fluidsynth.h>

#include "fweelin_core_dsp.h"

// Whacky FluidSynth center of pitchbend is 0x2000 not 0
#define FLUIDSYNTH_PITCHBEND_CENTER 0x2000

// Small class to encapsulate FluidSynth patches
class FluidPatch {
public:

  FluidPatch (int sfontid = 0, int bank = 0, int prog = 0, char *name = 0) :
    sfontid(sfontid), bank(bank), prog(prog) {};

  int sfontid,
    bank,
    prog;
  char *name;
};

class FluidPatches {
 public:
  FluidPatches() : 
    patches(0), numpatches(0), patchnum(0) {};

  // FluidSynth patch change support
  FluidPatch *patches;
  int numpatches, // Total # of patches
    patchnum; // Current patch #
};

// Integrated soft-synth based on libfluidsynth
class FluidSynthProcessor : public Processor, public EventListener {
  friend class Fweelin;

public:
  FluidSynthProcessor(Fweelin *app);
  virtual ~FluidSynthProcessor();

  virtual void process(char pre, nframes_t len, AudioBuffers *ab);

  sample_t *GetFluidOutput() { return outbuf; };
  FluidPatches *getPATCHES() { return &patches; };

  void ReceiveEvent(Event *ev, EventProducer *from);

  // Read new current patch from 'patches' and send patch change to synth
  void SendPatchChange();

  // Change patches by moving 'patchofs' patches forward or 'fontofs' soundfonts
  // forward.
  void AdjustPatch (int patchofs, int fontofs);

  // Enable/disable FluidSynth- if disabled, bypasses
  // processor stage to reduce CPU usage, but leaves memory allocated
  void SetEnable(char en) { 
    // Preprocess audio for smoothing
    //dopreprocess();
    this->enable = en;
  };

  char GetEnable() { return enable; };

 private:
  // Sets up our internal patch list based on loaded soundfonts
  void SetupPatches();

  // Left, right and mixed output buffers
  sample_t *leftbuf, 
    *rightbuf,
    *outbuf;

  // FluidSynth patch change support
  FluidPatches patches;

  // And fluidsynth variables..
  fluid_settings_t *settings;
  fluid_synth_t *synth;

  // Currently processing?
  char enable;
};

#endif
#endif
